@@ -1,9 +1,7 @@
-(import (only (gnu packages compression) gzip))
-
 (define-public create-file
   (make-process
    (name "create-file")
-   (outputs (list "/tmp/file.txt"))
+   (outputs (list "file.txt"))
    (run-time (complexity
               (space 20 mebibytes)
               (time  10)))
@@ -15,9 +13,9 @@
 (define-public compress-file
   (make-process
    (name "compress-file")
-   (packages (list gzip))
-   (inputs (list "/tmp/file.txt"))
-   (outputs (list "/tmp/file.txt.gz"))
+   (packages (list "gzip"))
+   (inputs (list "file.txt"))
+   (outputs (list "file.txt.gz"))
    (run-time (complexity
               (space 20 mebibytes)
               (time  10)))
