@@ -1,4 +1,4 @@
-;;; Copyright © 2019, 2020 Ricardo Wurmus <rekado@elephly.net>
+;;; Copyright © 2019, 2020, 2021 Ricardo Wurmus <rekado@elephly.net>
 ;;;
 ;;; This program is free software; you can redistribute it and/or modify it
 ;;; under the terms of the GNU General Public License as published by
@@ -14,8 +14,9 @@
 ;;; along with this program.  If not, see <http://www.gnu.org/licenses/>.
 
 (define-module (gwl cache)
+  #:use-module (gwl ui)
   #:use-module ((gwl processes)
-                #:select (process-inputs))
+                #:select (process-inputs process->script-arguments))
   #:use-module ((gwl workflows)
                 #:select (workflow-restrictions))
   #:use-module ((gwl workflows utils)
@@ -23,20 +24,22 @@
   #:use-module ((guix base32)
                 #:select (bytevector->base32-string))
   #:use-module ((rnrs bytevectors)
-                #:select (bytevector->u8-list
+                #:select (string->utf8
+                          bytevector->u8-list
                           u8-list->bytevector))
   #:use-module ((rnrs io ports)
                 #:select (get-bytevector-all))
   #:use-module ((gcrypt hash)
                 #:select (sha256))
-  #:use-module ((ice-9 iconv)
-                #:select (string->bytevector))
   #:use-module (ice-9 match)
   #:use-module (srfi srfi-1)
   #:use-module (srfi srfi-26)
   #:export (%cache-root
             %cache-delay
 
+            process->hash
+            process->hash-string
+
             make-process->cache-prefix
             cache!
             restore!))
@@ -47,7 +50,44 @@
 (define %cache-root (make-parameter "/tmp/gwl"))
 (define %cache-delay (make-parameter 0))
 
-(define (workflow->data-hashes workflow processes free-inputs-map make-script)
+(define (process->hash process scripts-table)
+  "Return a hash of the process.  The name of the script file already
+contains the hash of all its inputs, so we use that instead of the
+contents of the script, which we don't want to compute right now.
+Since the same scripts may be called with different arguments, we
+*also* need to include the process parameters in this computation.
+
+SCRIPTS-TABLE is a hash table of script file names for any given
+process."
+  (let ((script+arguments
+         (format #false "~a ~a"
+                 (hash-ref scripts-table process)
+                 (process->script-arguments process))))
+    (bytevector->u8-list
+     (sha256 (string->utf8 script+arguments)))))
+
+(define process->hash-string
+  (compose bytevector->base32-string
+           u8-list->bytevector
+           process->hash))
+
+(define (hash-input-file file-name)
+  "Return a hash representing the input file FILE-NAME.  We don't
+actually hash all the file, because that might be very expensive
+--- especially when done *every* time a workflow is supposed to be
+run.  We just hash the file name, the mtime and the file size and hope
+that's enough.
+
+We should probably have a cache for file hashes, so that hashing even
+large files can be acceptable as it only has to be done once."
+  (let ((st (stat file-name)))
+    (bytevector->u8-list
+     (sha256 (string->utf8 (format #f "~a~a~a"
+                                   file-name
+                                   (stat:mtime st)
+                                   (stat:size st)))))))
+
+(define (workflow->data-hashes workflow processes free-inputs-map scripts-table)
   "Return an alist associating each of the WORKFLOW's ordered list of
 PROCESSES with the hash of all the process scripts used to generate
 their outputs.  FREE-INPUTS-MAP is an alist of input names to file
@@ -57,41 +97,22 @@ names that must be considered when computing the hash."
     (filter-map (lambda (input)
                   (and=> (assoc-ref free-inputs-map input) first))
                 (process-inputs process)))
-  ;; XXX: We don't hash all of the inputs, because that might be very
-  ;; expensive --- especially when done *every* time a workflow is
-  ;; supposed to be run.  We just hash the file name, the mtime and
-  ;; the file size and hope that's enough.
-
-  ;; XXX: We should probably have a cache for file hashes, so that
-  ;; hashing even large files can be acceptable as it only has to be
-  ;; done once.
   (define input-hashes
     (map (match-lambda
            ((name file-name)
-            (let ((st (stat file-name)))
-              (cons file-name
-                    (bytevector->u8-list
-                     (sha256 (string->bytevector (format #f "~a~a~a"
-                                                         file-name
-                                                         (stat:mtime st)
-                                                         (stat:size st))
-                                                 "ISO-8859-1")))))))
+            (cons file-name (hash-input-file file-name))))
          free-inputs-map))
   ;; Compute hashes for chains of scripts.
   (define (kons process acc)
-    (let* ((script (make-script process #:workflow workflow))
-           (hash   (bytevector->u8-list
-                    (sha256 (call-with-input-file script get-bytevector-all)))))
-      (cons
-       (cons process
-             (append hash
-                     ;; Hash of mapped free inputs.
-                     (append-map (cut assoc-ref acc <>)
-                                 (process-free-inputs process))
-                     ;; All outputs of processes this one depends on.
-                     (append-map (cut assoc-ref acc <>)
-                                 (or (assoc-ref graph process) '()))))
-       acc)))
+    (cons (cons process
+                (append (process->hash process scripts-table)
+                        ;; Hash of mapped free inputs.
+                        (append-map (cut assoc-ref acc <>)
+                                    (process-free-inputs process))
+                        ;; All outputs of processes this one depends on.
+                        (append-map (cut assoc-ref acc <>)
+                                    (or (assoc-ref graph process) '()))))
+          acc))
   (map (match-lambda
          ((process . hashes)
           (cons process
@@ -101,7 +122,7 @@ names that must be considered when computing the hash."
        (fold kons input-hashes processes)))
 
 (define (make-process->cache-prefix workflow free-inputs-map ordered-processes
-                                    make-script)
+                                    scripts-table)
   "Return a procedure that takes a process and returns the cache
 prefix for its outputs."
   (let ((hashes (workflow->data-hashes workflow
@@ -111,7 +132,7 @@ prefix for its outputs."
                                                      (l (list l)))
                                                    ordered-processes)
                                        free-inputs-map
-                                       make-script)))
+                                       scripts-table)))
     (lambda (process)
       (and=> (assoc-ref hashes process)
              (cut string-append (%cache-root) "/" <> "/")))))
@@ -120,6 +141,17 @@ prefix for its outputs."
   (and=> (stat file #f)
          (lambda (st) (eq? 'directory (stat:type st)))))
 
+(define (link-or-symlink source target)
+  ;; Cross-device links don't work, of course, so we try symlinking
+  ;; on error.
+  (catch 'system-error
+    (lambda ()
+      (link source target))
+    (lambda (key . args)
+      (if (string=? "link" (car args))
+          (symlink source target)
+          (throw key args)))))
+
 (define (cache! file cache-prefix)
   "Cache FILE by linking it to the directory CACHE-PREFIX."
   (unless (directory? file)
@@ -129,7 +161,9 @@ prefix for its outputs."
     (let ((cached-file (string-append cache-prefix file)))
       (when (file-exists? cached-file)
         (delete-file cached-file))
-      (link file cached-file))))
+      (log-event 'cache (G_ "Caching `~a' as `~a'~%")
+                 file cached-file)
+      (link-or-symlink file cached-file))))
 
 (define (restore! file cache-prefix)
   "Restore FILE from the cache at CACHE-PREFIX."
@@ -137,6 +171,6 @@ prefix for its outputs."
     (mkdir-p (dirname file))
     (when (file-exists? file)
       (delete-file file))
-    ;; TODO: cross-device links don't work, of course.  Copy?  Symlink?
-    ;; Make this configurable?
-    (link (string-append cache-prefix file) file)))
+    (log-event 'cache (G_ "Restoring `~a' to `~a'~%")
+               (string-append cache-prefix file) file)
+    (link-or-symlink (string-append cache-prefix file) file)))
