@@ -1,5 +1,5 @@
 ;;; Copyright © 2017, 2018 Roel Janssen <roel@gnu.org>
-;;; Copyright © 2018, 2019, 2020 Ricardo Wurmus <rekado@elephly.net>
+;;; Copyright © 2018, 2019, 2020, 2021 Ricardo Wurmus <rekado@elephly.net>
 ;;;
 ;;; This program is free software; you can redistribute it and/or modify it
 ;;; under the terms of the GNU General Public License as published by
@@ -16,43 +16,27 @@
 
 (define-module (gwl processes)
   #:use-module (oop goops)
+  #:use-module (gwl errors)
   #:use-module (gwl oop)
   #:use-module (gwl process-engines)
-  #:use-module ((guix monads)
-                #:select (mlet* mlet mwhen return))
-  #:use-module ((guix derivations)
-                #:select (derivation->output-path
-                          built-derivations
-                          derivation-input
-                          derivation-input-output-paths))
+  #:use-module (gwl packages)
   #:use-module ((guix profiles)
                 #:select
-                (manifest-search-paths
-                 packages->manifest
-                 profile-derivation))
+                (profile
+                 manifest-search-paths
+                 packages->manifest))
   #:use-module ((guix search-paths)
                 #:select
                 (search-path-specification->sexp))
-  #:use-module ((guix packages)
-                #:select (package?))
-  #:use-module ((gnu packages)
-                #:select (specification->package))
-  #:use-module ((gnu packages bash)
-                #:select (bash-minimal))
   #:use-module (guix gexp)
-  #:use-module ((guix store)
-                #:select
-                (%store-monad
-                 with-store
-                 run-with-store
-                 store-lift
-                 requisites))
   #:use-module ((guix modules)
                 #:select (source-module-closure))
   #:use-module (ice-9 format)
   #:use-module (ice-9 match)
   #:use-module (srfi srfi-1)
   #:use-module (srfi srfi-26)
+  #:use-module (srfi srfi-34)
+  #:use-module (srfi srfi-35)
   #:export (make-process
             process?
             process-name
@@ -77,6 +61,8 @@
             complexity-threads
 
             process->script
+            process->script-arguments
+            process->script-wrapper
 
             ;; Convenience functions
             kibibytes
@@ -104,7 +90,7 @@
             code-snippet-arguments
             code-snippet-code
 
-            procedure->gexp
+            compile-procedure
             containerize))
 
 ;;; Commentary:
@@ -176,6 +162,7 @@
   (name
    #:accessor process-name
    #:init-keyword #:name
+   #:implicit-concatenation? #t
    #:required? #t)
   (version
    #:accessor process-version
@@ -195,7 +182,7 @@
    #:init-value '()
    #:implicit-list? #t
    #:validator? (lambda (value)
-                  (every package? value))
+                  (every valid-package? value))
    #:transformer
    ;; TODO: the instance name is not be available at this point, so we
    ;; can't report the process name here.  We should move the
@@ -203,15 +190,15 @@
    (lambda (instance value)
      (map (match-lambda
             ((and (? string?) spec)
-             (catch #t
-               (lambda ()
-                 (specification->package spec))
-               (lambda _
-                 (error (format #f "no such package: ~a~%" spec)))))
-            ((and (? package?) pkg)
+             (lookup-package spec))
+            ((and (? valid-package?) pkg)
              pkg)
             (x
-             (error (format #f "must provide package value or string: ~a~%" x))))
+             (raise
+              (condition
+               (&gwl-type-error
+                (expected-type (list "<package>" "<inferior-package>" "<string>"))
+                (actual-value x))))))
           value)))
   (inputs
    #:accessor process-raw-inputs
@@ -233,6 +220,11 @@
    #:init-keyword #:run-time
    #:init-value #f
    #:validator complexity?)
+  (values
+   #:accessor process-values
+   #:init-keyword #:values
+   #:init-value '()
+   #:implicit-list? #t)
   (procedure
    #:accessor process-procedure
    #:init-keyword #:procedure
@@ -280,8 +272,19 @@
                               ((empty-field)
                                (syntax-violation #f "process: Empty field" #'empty-field))
                               ((key value)
+                               ;; No valid fields are keywords.
+                               ;; Signal this error early instead of
+                               ;; constructing an invalid let*.
+                               (when (keyword? (syntax->datum #'key))
+                                 (syntax-violation #f "process: Invalid field name" #'key))
                                #'(key value))
                               ((key values ...)
+                               ;; No valid fields are keywords.
+                               ;; Signal this error early instead of
+                               ;; constructing an invalid let*.
+                               (when (keyword? (syntax->datum #'key))
+                                 (syntax-violation #f "process: Invalid field name" #'key))
+
                                ;; XXX: This is a crude way to allow
                                ;; for definitions inside of field
                                ;; values.
@@ -337,18 +340,16 @@
 (define (process->env process)
   "Return an alist of environment variable names to values of fields
 of PROCESS."
-  `(("_GWL_PROCESS_NAME" .
-     ,(process-name process))
+  `(("_GWL_PROCESS_NAME" . name)
     ("_GWL_PROCESS_SYNOPSIS" .
      ,(or (process-synopsis process) ""))
     ("_GWL_PROCESS_DESCRIPTION" .
      ,(or (process-description process) ""))
-    ("_GWL_PROCESS_INPUTS" .
-     ,(format #f "~{~a~^ ~}" (process-inputs process)))
+    ("_GWL_PROCESS_INPUTS" . inputs)
     ("_GWL_PROCESS_OUTPUT_PATH" .
      ,(or (process-output-path process) ""))
-    ("_GWL_PROCESS_OUTPUTS" .
-     ,(format #f "~{~a~^ ~}" (process-outputs process)))
+    ("_GWL_PROCESS_OUTPUTS" . outputs)
+    ("_GWL_PROCESS_VALUES" . values)
     ("_GWL_PROCESS_COMPLEXITY_THREADS" .
      ,(or (and=> (process-threads process) number->string) ""))
     ("_GWL_PROCESS_COMPLEXITY_SPACE" .
@@ -356,54 +357,53 @@ of PROCESS."
     ("_GWL_PROCESS_COMPLEXITY_TIME" .
      ,(or (and=> (process-time process) number->string) ""))))
 
+(define-syntax-rule (snippet-caller code-converter command)
+  (lambda (process code)
+    `(begin
+       (use-modules (srfi srfi-1))
+       (for-each (lambda (pair)
+                   (setenv (car pair)
+                           (let ((value (cdr pair)))
+                             (if (symbol? value)
+                                 (format #false "~a"
+                                         (assoc-ref #{ %gwl process-arguments}# value))
+                                 value))))
+                 ',(process->env process))
+       (let ((retval (apply system*
+                            (append command
+                                    (code-converter (string-join (map (lambda (val)
+                                                                        (if (list? val)
+                                                                            (format #f "~{~a~^ ~}" val)
+                                                                            (format #f "~a" val)))
+                                                                      (list ,@code))
+                                                                 ""))))))
+         (or (zero? retval) (exit retval))))))
+
 (define language-python
   (make <language>
     #:name 'python
-    #:call (lambda (process code)
-             #~(begin
-                 (for-each (lambda (pair)
-                             (setenv (car pair) (cdr pair)))
-                           '#$(process->env process))
-                 (let ((retval (system* "python3" "-c" #$code)))
-                   (or (zero? retval) (exit retval)))))))
+    #:call (snippet-caller list '("python3" "-c"))))
 
 (define language-r
   (make <language>
     #:name 'R
-    #:call (lambda (process code)
-             (let ((args (append-map (lambda (line)
-                                       (list "-e" line))
-                                     (filter (negate string-null?)
-                                             (string-split code #\newline)))))
-               #~(begin
-                   (for-each (lambda (pair)
-                               (setenv (car pair) (cdr pair)))
-                             '#$(process->env process))
-                   (let ((retval (apply system* "Rscript" '#$args)))
-                     (or (zero? retval) (exit retval))))))))
+    #:call (snippet-caller
+            (lambda (code)
+              (append-map (lambda (line)
+                            (list "-e" line))
+                          (filter (negate string-null?)
+                                  (string-split code #\newline))))
+            '("Rscript"))))
 
 (define language-bash
   (make <language>
     #:name 'bash
-    #:call (lambda (process code)
-             #~(begin
-                 (for-each (lambda (pair)
-                             (setenv (car pair) (cdr pair)))
-                           '#$(process->env process))
-                 (let ((retval (system* "bash" "-c" #$code)))
-                   (or (zero? retval) (exit retval)))))))
+    #:call (snippet-caller list '("bash" "-c"))))
 
 (define language-sh
   (make <language>
     #:name 'sh
-    #:call (lambda (process code)
-             #~(begin
-                 (for-each (lambda (pair)
-                             (setenv (car pair) (cdr pair)))
-                           '#$(process->env process))
-                 (let ((retval (system* "/bin/sh"
-                                        "-c" #$code)))
-                   (or (zero? retval) (exit retval)))))))
+    #:call (snippet-caller list '("/bin/sh" "-c"))))
 
 (define languages
   (list language-sh
@@ -420,7 +420,7 @@ of PROCESS."
    #:accessor code-snippet-arguments)
   (code
    #:init-keyword #:code
-   #:accessor code-snippet-code-listing))
+   #:accessor code-snippet-code))
 
 ;; Code snippets should be evaluating to themselves for more
 ;; convenient Wisp use.
@@ -432,23 +432,16 @@ of PROCESS."
 (define (code-snippet? thing)
   (is-a? thing <code-snippet>))
 
-(define (code-snippet-code code-snippet)
-  "Return the code listing as a string."
-  (string-join (map (lambda (val)
-                      (if (list? val)
-                          (format #f "~{~a~^ ~}" val)
-                          (format #f "~a" val)))
-                    (code-snippet-code-listing code-snippet))
-               ""))
-
 (define (code-snippet language arguments code)
   (make <code-snippet>
     #:language language
     #:arguments arguments
     #:code code))
 
-(define (procedure->gexp process)
-  "Transform the procedure of PROCESS to a G-expression."
+(define (compile-procedure process)
+  "Transform the procedure of PROCESS to an expression that can be
+written to a file."
+  ;; TODO: replace with normalize-file-name in utils?
   (define (sanitize-path path)
     (string-join (delete ".." (string-split path #\/))
                  "/"))
@@ -457,7 +450,6 @@ of PROCESS."
     ((? list? s) s)
     ((? code-snippet? snippet)
      (let* ((name (code-snippet-language snippet))
-            (arguments (code-snippet-arguments snippet))
             (code (code-snippet-code snippet))
             (call (or (and=> (find (lambda (lang)
                                      (eq? name (language-name lang)))
@@ -465,28 +457,30 @@ of PROCESS."
                              language-call)
                       ;; There is no pre-defined way to execute the
                       ;; snippet.  Use generic approach.
-                      (lambda (process code)
-                        #~(begin
-                            (for-each (lambda (pair)
-                                        (setenv (car pair) (cdr pair)))
-                                      '#$(process->env process))
-                            (apply system*
-                                   (string-append (getenv "_GWL_PROFILE")
-                                                  #$(sanitize-path (symbol->string name)))
-                                   '#$(append arguments (list code))))))))
+                      (snippet-caller list
+                                      (list
+                                       (string-append (getenv "_GWL_PROFILE")
+                                                      #$(sanitize-path (symbol->string name))))))))
        (call process code)))
     (whatever (error (format #f "unsupported procedure: ~a\n" whatever)))))
 
 
-(define* (containerize exp #:key inputs outputs)
-  "Wrap EXP, an S-expression or G-expression, in a G-expression that
-causes EXP to be run in a container where the provided INPUTS are
-available.  OUTPUTS are copied outside of the container."
+(define* (script-modules #:optional containerize?)
+  (if containerize?
+      '((gnu build accounts)
+        (gnu build linux-container)
+        (gnu system file-systems)
+        (guix build utils)
+        (guix search-paths))
+      '((guix search-paths))))
+
+(define* (containerize script #:key inputs outputs)
+  "Call SCRIPT, a program-file object, in a G-expression that sets up
+a container where the provided INPUTS are available (in addition to
+all store items needed for execution).  OUTPUTS are copied outside of
+the container."
   (with-imported-modules (source-module-closure
-                          '((gnu build accounts)
-                            (gnu build linux-container)
-                            (gnu system file-systems)
-                            (guix build utils)))
+                          (script-modules 'container))
     #~(begin
         (use-modules (gnu build accounts)
                      (gnu build linux-container)
@@ -502,8 +496,7 @@ available.  OUTPUTS are copied outside of the container."
                        '(bind-mount read-only)))
             (check? #f)
             (create-mount-point? #t)))
-        (let* ((thunk (lambda () #$exp))
-               (pwd (getpw))
+        (let* ((pwd (getpw))
                (uid (getuid))
                (gid (getgid))
                (passwd (let ((pwd (getpwuid (getuid))))
@@ -522,12 +515,19 @@ available.  OUTPUTS are copied outside of the container."
                              (canonicalize-path ".") "/gwl" #t))
                       (map (lambda (location)
                              (location->file-system location location #f))
-                           '#$inputs))
+                           (append (call-with-input-file #$(references-file script) read)
+                                   '#$inputs)))
             (lambda ()
               (unless (file-exists? "/bin/sh")
                 (unless (file-exists? "/bin")
                   (mkdir "/bin"))
-                (symlink #$(file-append bash-minimal "/bin/sh") "/bin/sh"))
+                (symlink #$(file-append (bash-minimal) "/bin/sh") "/bin/sh"))
+
+              ;; Setup directory for temporary files.
+              (mkdir-p "/tmp")
+              (for-each (lambda (var)
+                          (setenv var "/tmp"))
+                        '("TMPDIR" "TEMPDIR"))
 
               ;; Create a dummy /etc/passwd to satisfy applications that demand
               ;; to read it.
@@ -536,7 +536,7 @@ available.  OUTPUTS are copied outside of the container."
               (write-passwd (list passwd))
               (write-group groups)
 
-              (thunk)
+              (system* #$script)
 
               ;; Copy generated files to final directory.
               (for-each (lambda (output)
@@ -551,18 +551,23 @@ available.  OUTPUTS are copied outside of the container."
 ;;; ADDITIONAL FUNCTIONS
 ;;; ---------------------------------------------------------------------------
 
-(define (process-inputs process)
+(define* (process-inputs process #:optional with-tags?)
   "Return the plain values of all inputs of PROCESS, without any
-keyword tags."
-  (remove keyword? (process-raw-inputs process)))
+keyword tags if WITH-TAGS? is #FALSE or missing."
+  (if with-tags?
+      (process-raw-inputs process)
+      (remove keyword? (process-raw-inputs process))))
 
-(define (process-outputs proc)
-  "Return the output location(s) of process PROC."
+(define* (process-outputs proc #:optional with-tags?)
+  "Return the output location(s) of process PROC, without any keyword
+tags if WITH-TAGS? is #FALSE or missing."
   (let* ((root (process-output-path proc))
          (mangle (if root
                      (cut string-append root "/" <>)
                      identity)))
-    (map mangle (remove keyword? (process-outputs* proc)))))
+    (if with-tags?
+        (map mangle (process-outputs* proc))
+        (map mangle (remove keyword? (process-outputs* proc))))))
 
 (define (process-takes-available process)
   "Returns #T when the data inputs of the PROCESS exist."
@@ -575,81 +580,91 @@ keyword tags."
 ;;; DERIVATIONS AND SCRIPTS FUNCTIONS
 ;;; ---------------------------------------------------------------------------
 
-(define* (process->script engine #:key containerize?)
-  "Build a procedure that transforms the process PROCESS into a script
-and returns its location."
-  (lambda* (process #:key workflow (input-files '()))
-    (let* ((name (process-full-name process))
-           (exp (procedure->gexp process))
-           (make-wrapper (process-engine-wrapper engine))
-           (packages (cons bash-minimal
-                           (process-packages process)))
-           (manifest (packages->manifest packages))
-           (search-paths (delete-duplicates
-                          (map search-path-specification->sexp
-                               (manifest-search-paths manifest))))
-           (out (process-output-path process))
-           ;; TODO: with-imported-modules does not support nesting!
-           ;; So we need to list all the modules right here, or else
-           ;; containerization or search path magic simply won't work
-           ;; at runtime on all targets.
-           (modules (if containerize?
-                        '((gnu build accounts)
-                          (gnu build linux-container)
-                          (gnu system file-systems)
-                          (guix build utils)
-                          (guix search-paths))
-                        '((guix search-paths))))
-           (drv
-            (mlet* %store-monad
-                ((profile (profile-derivation manifest))
-                 (exp* -> (with-imported-modules (source-module-closure modules)
-                            #~(begin
-                                #$@(if (null? packages) '()
-                                       `((use-modules (guix search-paths))
-                                         (set-search-paths (map sexp->search-path-specification
-                                                                ',search-paths)
-                                                           (cons ,profile
-                                                                 ',packages))))
-                                #$(if out `(setenv "out" ,out) "")
-                                (setenv "_GWL_PROFILE" #$profile)
-                                #$exp)))
-                 (lowered (lower-gexp exp*))
-                 (inputs -> (cons* (derivation-input profile)
-                                   (lowered-gexp-guile lowered)
-                                   (lowered-gexp-inputs lowered)))
-                 (_ (built-derivations inputs))
-                 (items -> (append (append-map derivation-input-output-paths inputs)
-                                   (lowered-gexp-sources lowered)))
-                 (closure ((store-lift requisites) items))
-                 ;; Build everything
-                 (built (built-derivations closure))
-                 (script -> (if containerize?
-                                (containerize exp*
-                                              #:inputs
-                                              (append closure
-                                                      ;; Data inputs
-                                                      (process-inputs process)
-                                                      ;; Files mapped to free inputs
-                                                      input-files)
-                                              #:outputs
-                                              (process-outputs process))
-                                exp*)))
-              (gexp->script (string-append "gwl-" name ".scm") script))))
-      (with-store store
-        (run-with-store store
-          (mlet* %store-monad ((drv drv)
-                               (built (built-derivations (list drv))))
-            (if make-wrapper
-                (mlet* %store-monad
-                    ((wrap (gexp->derivation
-                            (string-append "gwl-launch-" name ".scm")
-                            (make-wrapper process
-                                          (derivation->output-path drv)
-                                          #:workflow workflow)))
-                     (built (built-derivations (list wrap))))
-                  (return (derivation->output-path wrap)))
-                (return (derivation->output-path drv)))))))))
+;; Taken from (gnu services base)
+(define* (references-file item #:optional (name "references"))
+  "Return a file that contains the list of references of ITEM."
+  (if (struct? item)                              ;lowerable object
+      (computed-file name
+                     (with-extensions (list (lookup-package "guile-gcrypt")) ;for store-copy
+                       (with-imported-modules (source-module-closure
+                                               '((guix build store-copy)))
+                         #~(begin
+                             (use-modules (guix build store-copy))
+
+                             (call-with-output-file #$output
+                               (lambda (port)
+                                 (write (map store-info-item
+                                             (call-with-input-file "graph"
+                                               read-reference-graph))
+                                        port))))))
+                     #:options `(#:local-build? #false
+                                 #:references-graphs (("graph" ,item))))
+      (plain-file name "()")))
+
+(define (process->script-arguments process)
+  `((inputs  . ,(process-inputs process #:with-tags))
+    (outputs . ,(process-outputs process #:with-tags))
+    (values  . ,(process-values process))
+    (name    . ,(process-name process))))
+
+(define* (process->script process
+                          #:key
+                          containerize?
+                          workflow
+                          (input-files '()))
+  "Return a lowerable object for the script that will execute the
+PROCESS."
+  (let* ((name         (process-full-name process))
+         (packages     (cons (bash-minimal)
+                             (process-packages process)))
+         (manifest     (packages->manifest packages))
+         (profile      (profile (content manifest)))
+         (search-paths (delete-duplicates
+                        (map search-path-specification->sexp
+                             (manifest-search-paths manifest))))
+         (out          (process-output-path process))
+         (exp
+          (with-imported-modules (source-module-closure (script-modules))
+            #~(begin
+                #$@(if (null? packages) '()
+                       `((use-modules (guix search-paths))
+                         (set-search-paths (map sexp->search-path-specification
+                                                ',search-paths)
+                                           (cons ,profile
+                                                 ',packages))))
+                #$(if out `(setenv "out" ,out) "")
+                (setenv "_GWL_PROFILE" #$profile)
+                (use-modules (ice-9 match))
+                (match (command-line)
+                  ((_ (= (lambda (s) (call-with-input-string s read))
+                         #{ %gwl process-arguments}#) . rest)
+                   #$(compile-procedure process))))))
+         (script
+          (program-file (string-append "gwl-" name ".scm")
+                        exp
+                        #:guile (default-guile)))
+         (computed-script
+          (if containerize?
+              (program-file (string-append "gwl-" name "-container.scm")
+                            (containerize script
+                                          #:inputs
+                                          (append
+                                           ;; Data inputs
+                                           (process-inputs process)
+                                           ;; Files mapped to free inputs
+                                           input-files)
+                                          #:outputs
+                                          (process-outputs process))
+                            #:guile (default-guile))
+              script)))
+    computed-script))
+
+(define* (process->script-wrapper process
+                                  #:key make-wrapper workflow scripts-table)
+  (make-wrapper process
+                (hash-ref scripts-table process)
+                #:workflow workflow
+                #:scripts-table scripts-table))
 
 ;;; ---------------------------------------------------------------------------
 ;;; CONVENIENCE FUNCTIONS
