@@ -1,5 +1,5 @@
 ;;; Copyright © 2016, 2017, 2018 Roel Janssen <roel@gnu.org>
-;;; Copyright © 2018, 2019, 2020 Ricardo Wurmus <rekado@elephly.net>
+;;; Copyright © 2018, 2019, 2020, 2021 Ricardo Wurmus <rekado@elephly.net>
 ;;;
 ;;; This program is free software; you can redistribute it and/or modify it
 ;;; under the terms of the GNU General Public License as published by
@@ -15,12 +15,27 @@
 ;;; along with this program.  If not, see <http://www.gnu.org/licenses/>.
 
 (define-module (gwl workflows)
+  #:use-module (gwl config)
   #:use-module (gwl cache)
   #:use-module (gwl oop)
+  #:use-module (gwl packages)
   #:use-module (gwl processes)
   #:use-module (gwl process-engines)
   #:use-module (gwl workflows execution-order)
   #:use-module (gwl workflows utils)
+  #:use-module (gwl ui)
+  #:use-module (guix monads)
+  #:use-module (guix store)
+  #:use-module (guix gexp)
+  #:use-module ((guix derivations)
+                #:select (built-derivations
+                          derivation-outputs
+                          derivation-output-path))
+  #:use-module ((guix ui)
+                #:select (build-notifier))
+  #:use-module ((guix status)
+                #:select (with-status-verbosity))
+
   #:use-module (ice-9 match)
   #:use-module (ice-9 format)
   #:use-module (srfi srfi-1)
@@ -42,6 +57,13 @@
 
             print-workflow-record
 
+            compute-workflow
+            computed-workflow?
+            computed-workflow-workflow
+            computed-workflow-script-proc
+            computed-workflow-ordered-processes
+            computed-workflow-cache-prefix-proc
+
             workflow-free-inputs
             workflow-run-order
             workflow-prepare
@@ -133,8 +155,18 @@ Use \"processes\" to specify process dependencies.~%"))
                             ((empty-field)
                              (syntax-violation #f "workflow: Empty field" #'empty-field))
                             ((key value)
+                             ;; No valid fields are keywords.
+                             ;; Signal this error early instead of
+                             ;; constructing an invalid let*.
+                             (when (keyword? (syntax->datum #'key))
+                               (syntax-violation #f "workflow: Invalid field name" #'key))
                              #'(key value))
                             ((key values ...)
+                             ;; No valid fields are keywords.
+                             ;; Signal this error early instead of
+                             ;; constructing an invalid let*.
+                             (when (keyword? (syntax->datum #'key))
+                               (syntax-violation #f "workflow: Invalid field name" #'key))
                              ;; XXX: This is a crude way to allow
                              ;; for definitions inside of field
                              ;; values.
@@ -155,6 +187,114 @@ Use \"processes\" to specify process dependencies.~%"))
                                  #`((symbol->keyword 'name) name))))
                             #'(fields ...))))))))
 
+
+(define-immutable-record-type <computed-workflow>
+  (make-computed-workflow workflow
+                          script-proc
+                          ordered-processes
+                          cache-prefix-proc)
+  computed-workflow?
+  (workflow           computed-workflow-workflow)           ; <workflow>
+  (script-proc        computed-workflow-script-proc)        ; procedure taking a <process>
+  (ordered-processes  computed-workflow-ordered-processes)  ; list (of lists) of <process>
+  (cache-prefix-proc  computed-workflow-cache-prefix-proc)) ; procedure taking a <process>
+
+(define* (compute-workflow workflow
+                           #:key
+                           engine
+                           (inputs '())
+                           (parallel? #true)
+                           containerize?)
+  "Return a <computed-workflow> record value."
+  (define inputs-map (inputs->map inputs))
+  (define inputs-map-with-extra-files
+    (prepare-inputs workflow inputs-map))
+  (define ordered-processes
+    (workflow-run-order workflow #:parallel? parallel?))
+
+  (define (scripts-by-process)
+    (let ((h (make-hash-table)))
+      (for-each (lambda (process)
+                  (log-event 'debug
+                             (G_ "Computing script for process `~a'~%")
+                             (process-name process))
+                  (hash-set! h process
+                             (process->script process
+                                              #:containerize? containerize?
+                                              #:workflow workflow
+                                              #:input-files
+                                              (lset-intersection
+                                               string=?
+                                               (map second inputs-map-with-extra-files)
+                                               (process-inputs process)))))
+                (workflow-processes workflow))
+      h))
+  (define (wrapper-scripts-by-process scripts-table)
+    (let ((make-wrapper (process-engine-wrapper engine)))
+      (if make-wrapper
+          (let ((h (make-hash-table)))
+            (for-each (lambda (process)
+                        (log-event 'debug
+                                   (G_ "Computing wrapper script for process `~a'~%")
+                                   (process-name process))
+                        (hash-set! h process
+                                   (process->script-wrapper process
+                                                            #:make-wrapper make-wrapper
+                                                            #:workflow workflow
+                                                            #:scripts-table scripts-table)))
+                      (workflow-processes workflow))
+            h)
+          scripts-table)))
+  (define (script-files-by-process scripts-table)
+    (let* ((h (make-hash-table))
+           (processes (workflow-processes workflow))
+           (scripts (map-in-order (lambda (process)
+                                    (hash-ref scripts-table process))
+                                  processes))
+           (file-names
+            (parameterize ((%guile-for-build (default-guile-derivation)))
+              (with-status-verbosity (%config 'verbosity)
+                (with-build-handler (build-notifier #:verbosity (%config 'verbosity))
+                  (run-with-store (inferior-store)
+                    (mlet* %store-monad
+                        ((drvs (mapm/accumulate-builds lower-object scripts))
+                         (%    (built-derivations drvs)))
+                      (return
+                       (map (match-lambda
+                              (((_ . output) . rest)
+                               (derivation-output-path output)))
+                            (map derivation-outputs drvs))))))))))
+      (for-each (lambda (process script-file)
+                  (log-event 'debug
+                             (G_ "Recording script file for process `~a'~%")
+                             (process-name process))
+                  (hash-set! h process script-file))
+                processes
+                file-names)
+      h))
+  (define plain-scripts-table
+    (scripts-by-process))
+  (define plain-script-files-table
+    (script-files-by-process plain-scripts-table))
+  (define wrapper-scripts-table
+    (wrapper-scripts-by-process plain-scripts-table))
+  (define wrapper-script-files-table
+    (script-files-by-process wrapper-scripts-table))
+  (define (script-for-process process)
+    (hash-ref wrapper-script-files-table process))
+
+  (define process->cache-prefix
+    (make-process->cache-prefix workflow
+                                inputs-map-with-extra-files
+                                ordered-processes
+                                wrapper-script-files-table))
+
+  (make-computed-workflow workflow
+                          script-for-process
+                          ordered-processes
+                          process->cache-prefix))
+
+
 (define-syntax graph
   (lambda (x)
     (syntax-case x (->)
@@ -242,14 +382,14 @@ contains lists of processes that can be executed in parallel."
         (error (format (current-error-port)
                        "error: Cannot determine process execution order.~%")))))
 
-(define (workflow-kons workflow proc)
+(define (workflow-kons proc)
   "Construct a procedure from the single-argument procedure PROC that
-can be used in a fold over the WORKFLOW's processes."
+can be used in a fold over a WORKFLOW's processes."
   (lambda (item acc)
     (match item
       ((? list?)
        (fold (lambda (process res)
-               (cons (proc process #:workflow workflow) res))
+               (cons (proc process) res))
              acc
              ;; By reversing the order of the processes
              ;; in STEP we keep the output order the same
@@ -259,17 +399,33 @@ can be used in a fold over the WORKFLOW's processes."
 
 (define* (workflow-prepare workflow engine
                            #:key
+                           (inputs '())
                            (parallel? #t)
                            containerize?)
   "Print scripts to be run for WORKFLOW given ENGINE."
-  (define ordered-processes
-    (workflow-run-order workflow #:parallel? parallel?))
+  (define computed-workflow
+    (guard (condition
+            ((missing-inputs-condition? condition)
+             (log-event 'error
+                        (G_ "Missing inputs: ~{~%  * ~a~}.~%Provide them with --input=NAME=FILE.~%")
+                        (missing-inputs-files condition))
+             (exit 1)))
+      (log-event 'info (G_ "Computing workflow `~a'...~%")
+                 (workflow-name workflow))
+      (compute-workflow workflow
+                        #:engine engine
+                        #:inputs inputs
+                        #:parallel? parallel?
+                        #:containerize? containerize?)))
+  (define script-for-process
+    (computed-workflow-script-proc computed-workflow))
   (for-each (lambda (command)
               (display command) (newline))
-            (reverse (fold (workflow-kons
-                            workflow
-                            (process->script engine #:containerize? containerize?))
-                           '() ordered-processes))))
+            (delete-duplicates
+             (reverse (fold (workflow-kons script-for-process)
+                            '()
+                            (computed-workflow-ordered-processes computed-workflow)))
+             string=?)))
 
 (define (inputs->map inputs)
   "Given a list of strings INPUTS of the format \"a=b\" or just \"a\",
@@ -306,7 +462,7 @@ of missing files."
   (match unspecified-inputs
     (()
      ;; verify input files
-     (match (filter (negate file-exists?) input-files)
+     (match (remove file-exists? input-files)
        (()
         ;; Link all mapped input files to their target locations
         ;; TODO: ensure that target directories exist.
@@ -345,25 +501,23 @@ to existing files.
 
 When CONTAINERIZE? is #T build a process script that spawns a
 container."
-  (define inputs-map (inputs->map inputs))
-  (define inputs-map-with-extra-files
+  (define computed-workflow
     (guard (condition
             ((missing-inputs-condition? condition)
-             (format (current-error-port)
-                     "Missing inputs: ~{~%  * ~a~}.~%Provide them with --input=NAME=FILE.~%"
-                     (missing-inputs-files condition))
+             (log-event 'error
+                        (G_ "Missing inputs: ~{~%  * ~a~}.~%Provide them with --input=NAME=FILE.~%")
+                        (missing-inputs-files condition))
              (exit 1)))
-      (prepare-inputs workflow inputs-map)))
-  (define ordered-processes
-    (workflow-run-order workflow #:parallel? parallel?))
-  (define make-script
-    (process->script engine #:containerize? containerize?))
+      (log-event 'info (G_ "Computing workflow `~a'...~%")
+                 (workflow-name workflow))
+      (compute-workflow workflow
+                        #:engine engine
+                        #:inputs inputs
+                        #:parallel? parallel?
+                        #:containerize? containerize?)))
   (define runner (process-engine-runner engine))
   (define process->cache-prefix
-    (make-process->cache-prefix workflow
-                                inputs-map-with-extra-files
-                                ordered-processes
-                                make-script))
+    (computed-workflow-cache-prefix-proc computed-workflow))
   (define cached?
     (if force?
         (const #f)
@@ -374,19 +528,21 @@ container."
                           (file-exists?
                            (string-append cache-prefix out)))
                         (process-outputs process)))))))
-  (define* (run process #:key workflow)
+  (define script-for-process
+    (computed-workflow-script-proc computed-workflow))
+  (define (run process)
     (let ((cache-prefix (process->cache-prefix process)))
       (if (cached? process)
           (if dry-run?
-              (format (current-error-port)
-                      "Would skip process \"~a\" (cached at ~a).~%"
-                      (process-name process)
-                      cache-prefix)
+              (log-event 'execute
+                         (G_ "Would skip process \"~a\" (cached at ~a).~%")
+                         (process-name process)
+                         cache-prefix)
               (begin
-                (format (current-error-port)
-                        "Skipping process \"~a\" (cached at ~a).~%"
-                        (process-name process)
-                        cache-prefix)
+                (log-event 'execute
+                           (G_ "Skipping process \"~a\" (cached at ~a).~%")
+                           (process-name process)
+                           cache-prefix)
                 ;; TODO: mount the cache directory in the container
                 ;; if containerized.  Otherwise link files from
                 ;; cache to expected location.
@@ -394,27 +550,22 @@ container."
                           (process-outputs process))))
 
           ;; Not cached: execute the process!
-          (let* ((script
-                  (make-script process
-                               #:workflow workflow
-                               #:input-files
-                               (lset-intersection
-                                string=?
-                                (map second inputs-map-with-extra-files)
-                                (process-inputs process))))
-                 (command (append runner (list script))))
+          (let* ((built-script (script-for-process process))
+                 (command (runner (list built-script
+                                        (format #false "'~S'"
+                                                (process->script-arguments process))))))
             (if dry-run?
-                (format (current-error-port)
-                        "Would execute: ~{~a ~}~%" command)
+                (log-event 'execute
+                           (G_ "Would execute: ~{~a ~}~%") command)
                 (begin
-                  (format (current-error-port)
-                          "Executing: ~{~a ~}~%" command)
+                  (log-event 'execute
+                             (G_ "Executing: ~{~a ~}~%") command)
                   (let ((retval (apply system* command)))
                     (unless (zero? retval)
-                      (format (current-error-port)
-                              "error: process `~a' failed with return value ~a.~%"
-                              (process-name process)
-                              retval)
+                      (log-event 'error
+                                 (G_ "error: process `~a' failed with return value ~a.~%")
+                                 (process-name process)
+                                 retval)
                       (exit retval)))
                   ;; Wait before generated files are accessed.
                   ;; This may be needed for distributed file
@@ -427,15 +578,16 @@ container."
                                                         output
                                                         (string-append (getcwd) "/" output))))
                                 (unless (file-exists? canonical-name)
-                                  (format (current-error-port)
-                                          "error: process `~a' failed to produce output ~a.~%"
-                                          (process-name process)
-                                          output)
+                                  (log-event 'error
+                                             (G_ "process `~a' failed to produce output ~a.~%")
+                                             (process-name process)
+                                             output)
                                   (exit 1))))
                             (process-outputs process))
 
                   ;; Link files to the cache.
                   (for-each (cut cache! <> cache-prefix)
                             (process-outputs process))))))))
-  (fold (workflow-kons workflow run)
-        '() ordered-processes))
+  (fold (workflow-kons run)
+        '()
+        (computed-workflow-ordered-processes computed-workflow)))
