@@ -1,5 +1,5 @@
 ;;; Copyright © 2016, 2017  Roel Janssen <roel@gnu.org>
-;;; Copyright © 2019 Ricardo Wurmus <rekado@elephly.net>
+;;; Copyright © 2019, 2020, 2021 Ricardo Wurmus <rekado@elephly.net>
 ;;;
 ;;; This program is free software: you can redistribute it and/or
 ;;; modify it under the terms of the GNU Affero General Public License
@@ -16,6 +16,7 @@
 ;;; <http://www.gnu.org/licenses/>.
 
 (define-module (gwl www pages)
+  #:use-module (gwl config)
   #:use-module (srfi srfi-1)
   #:use-module (srfi srfi-26)
   #:use-module (ice-9 match)
@@ -30,14 +31,18 @@
     ("/community"  "Community")))
 
 (define (page-partial-main-menu request-path)
-  `(ul ,(map (match-lambda
-               (((? (cut string=? request-path <>)) label)
-                `(li (@ (class "active")) ,label))
-               ((url label)
-                `(li (a (@ (href ,url)) ,label))))
-             pages)))
+  `(ul ,@(map (match-lambda
+                (((? (lambda (i)
+                       (and (string? i)
+                            (string? request-path)
+                            (string=? request-path i)))) label)
+                 `((li (@ (class "active")) ,label)))
+                ((url label)
+                 `((li (a (@ (href ,url)) ,label))))
+                (_ '()))
+              pages)))
 
-(define* (page-root-template title request-path content-tree #:key (dependencies '()))
+(define (page-root-template title request-path content-tree)
   `((html (@ (lang "en"))
      (head
       (title ,(string-append page-title-prefix title))
@@ -45,11 +50,6 @@
       (link (@ (rel "icon")
                (type "image/x-icon")
                (href "/static/favicon.ico")))
-      ,(if (memq 'highlight dependencies)
-         `((link (@ (rel "stylesheet") (href "/static/highlight/styles/github.css")))
-           (script (@ (src "/static/highlight/highlight.js")) "")
-           (script (@ (type "text/javascript")) "hljs.initHighlightingOnLoad();"))
-         `())
       (link (@ (rel "stylesheet")
                (href "/static/css/main.css")
                (type "text/css")
@@ -65,5 +65,5 @@
                    (class "width-control"))
                 ,content-tree)
            (div (@ (id "footer"))
-                (p (a (@ (href "https://git.sv.gnu.org/cgit/gwl.git"))
+                (p (a (@ (href ,%gwl-git-browse-url))
                       "Download the source code of this page") ".")))))))
