@@ -1,4 +1,4 @@
-;;; Copyright © 2019, 2020 Ricardo Wurmus <rekado@elephly.net>
+;;; Copyright © 2019, 2020, 2021 Ricardo Wurmus <rekado@elephly.net>
 ;;;
 ;;; This program is free software; you can redistribute it and/or modify it
 ;;; under the terms of the GNU General Public License as published by
@@ -14,6 +14,7 @@
 ;;; along with this program.  If not, see <http://www.gnu.org/licenses/>.
 
 (define-module (test-workflows)
+  #:use-module (gwl errors)
   #:use-module (gwl processes)
   #:use-module (gwl workflows)
   #:use-module (gwl workflows utils)
@@ -22,6 +23,7 @@
   #:use-module (srfi srfi-1)
   #:use-module (srfi srfi-26)
   #:use-module (srfi srfi-34)
+  #:use-module (srfi srfi-35)
   #:use-module (srfi srfi-64))
 
 (test-begin "workflows")
@@ -247,4 +249,54 @@ make-workflow
     (prepare-inputs (wf3 prefix)
                     incomplete-inputs-map)))
 
+(test-equal "make-workflow rejects invalid field names 1/3"
+  '("workflow" (garbage))
+  (guard (c ((gwl-error? c)
+             (formatted-message-arguments c)))
+    (make-workflow
+     (name "anything")
+     (processes '())
+     (garbage 'this-is))))
+
+(test-equal "make-workflow rejects invalid field names 2/3"
+  "workflow: Invalid field name"
+  (catch 'syntax-error
+    (lambda ()
+      (test-read-eval-string (format #false "~s"
+                                     '(make-workflow
+                                       (name "anything")
+                                       (processes '())
+                                       (#:garbage 'this-is)))))
+    (lambda (key _ message . rest)
+      message)))
+
+(test-equal "make-workflow rejects invalid field names 3/3"
+  "workflow: Invalid field name"
+  (catch 'syntax-error
+    (lambda ()
+      (test-read-eval-string (format #false "~s"
+                                     '(make-workflow
+                                       (name "anything")
+                                       (processes '())
+                                       (#:garbage 'this-is 'still-garbage)))))
+    (lambda (key _ message . rest)
+      message)))
+
+#;
+(test-assert "make-workflow macro supports implicit lists"
+  (let ((wf (make-workflow
+             (name "anything")
+             (packages "hello" "coreutils"))))
+    (equal? '(this that whatever)
+            (workflow-packages wf))))
+
+#;
+(test-assert "make-process macro flattens implicit lists"
+  (let ((proc (make-process
+               (name "anything")
+               (procedure '(const #t))
+               (inputs '(this and) 'that '(whatever you say)))))
+    (equal? '(this and that whatever you say)
+            (process-inputs proc))))
+
 (test-end "workflows")
